using System;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections.Generic;
using BatchCat;

class Bib
{
    public int    id;
    public int    owningLibrary;
    public string data;
    public bool   suppress;
    public Bib ()                   { id = 0; }
    public Bib (int i)              { id = i; }
    public Bib (int i, string d)    { id = i; data = d; }
    public Bib (string d)           { id = 0; data = d; }
    public long Id ()               { return id; }
    public void Id (int i)          { id = i; }
    public void Id (string i)       { id = Convert.ToInt32(i); }
    public void Data (string d)     { data = d; }
    public void Suppress (bool s)   { suppress = s; }
    public void Suppress (string s) { suppress = s.Equals(@"yes"); }
}

class Item
{
    public int    id;
    public string barcode;
    private ClassItemRecord itemRecord;
    public Item ()                  { }
    public Item (ClassItemRecord i) { itemRecord = i; }
    public void Clear ()
    {
        id = 0;
        barcode = "";
        // Mandatory
        // NO NO NO!! itemRecord.HoldingID = 0;
        itemRecord.ItemID = 0;
        itemRecord.ItemTypeID = 0;
        itemRecord.PermLocationID = 0;
        // Optional
        itemRecord.AddItemToTop = false;
        itemRecord.Caption = "";
        itemRecord.Chron = "";
        itemRecord.CopyNumber = 0;
        itemRecord.Enumeration = "";
        itemRecord.FreeText = "";
        itemRecord.MediaTypeID = 0;
        itemRecord.PieceCount = 1;
        itemRecord.Price = "0";
        itemRecord.SpineLabel = "";
        itemRecord.TempLocationID = 0;
        itemRecord.TempTypeID = 0;
        itemRecord.Year = "";
    }
    // Required properties
    public void Id (int i)              { id = itemRecord.ItemID    = i; }
    public void HoldingId (int i)       { itemRecord.HoldingID      = i; }
    public void Id (string i)           { id = itemRecord.ItemID    = Convert.ToInt32(i); }
    public void HoldingID(string i)     { itemRecord.HoldingID      = Convert.ToInt32(i); }
    public void Type (string i)         { itemRecord.ItemTypeID     = Convert.ToInt32(i); }
    public void PermLocation(string i)  { itemRecord.PermLocationID = Convert.ToInt32(i); }
    // Optional properties
    public void Caption(string s)       { itemRecord.Caption        = s; }
    public void Chronology(string s)    { itemRecord.Chron          = s; }
    public void CopyNumber(string i)    { itemRecord.CopyNumber     = Convert.ToInt16(i); }
    public void Enumeration(string s)   { itemRecord.Enumeration    = s; }
    public void FreeText(string s)      { itemRecord.FreeText       = s; }
    public void MediaType(string i)     { itemRecord.MediaTypeID    = Convert.ToInt32(i); }
    public void Pieces(string i)        { itemRecord.PieceCount     = Convert.ToInt16(i); }
    public void Price(string s)         { itemRecord.Price          = s; }
    public void SpineLabel(string s)    { itemRecord.SpineLabel     = s; }
    public void TempLocation(string i)  { itemRecord.TempLocationID = Convert.ToInt32(i); }
    public void TempType(string i)      { itemRecord.TempTypeID     = Convert.ToInt32(i); }
    public void Year(string s)          { itemRecord.Year           = s; }
    // Barcode
    public void Barcode (string b)      { barcode = b; }
}

class Holding
{
    public int    id;
    public string data;
    public bool   suppress;
    public int    location;
    public Holding ()               { id = 0; }
    public void Id (int i)          { id = i; }
    public void Id (string i)       { id = Convert.ToInt32(i); }
    public void Data (string d)     { data = d; }
    public void Suppress (bool s)   { suppress = s; }
    public void Suppress (string s) { suppress = s.Equals(@"yes"); }
    public void Location (int i)    { location = i; }
    public void Location (string i) { location = Convert.ToInt32(i); }
}

public class RoboCatAgent
{

    // ------------------------------------------------------------------------- Instance variables --------

    // Streams
    private StreamReader    controlReader;
    private string          inputFile;
    private BinaryReader    inputReader;

    // Voyager
    private ClassBatchCat   batchCat;
    private string          voyagerINIDir = @"C:\Voyager";
    private string          userName = @"phoffman";
    private string          password;
    private int             catalogingLocation = 0;
    private Bib             bib;
    private Item            item;
    private Holding         holding;

    // State, etc.
    private bool            isRunning = true;
    private bool            isConnected = false;
    private bool            debug = false;
    private bool            echo = false;
    private bool            strict = false;
    private int             blockLevel = 0;
    private int             inError = 0;
    private string          command;

    // Misc
    private Encoding        utf8 = new UTF8Encoding(false, true);

    public RoboCatAgent()
    {
        controlReader = new StreamReader(Console.OpenStandardInput(), System.Text.Encoding.ASCII);
        batchCat = new ClassBatchCat();
        bib = new Bib();
        item = new Item(batchCat.cItem);
        holding = new Holding();
    }

    public RoboCatAgent(string cFile)
    {
        controlReader = new StreamReader(cFile, System.Text.Encoding.ASCII);
        batchCat = new ClassBatchCat();
        bib = new Bib();
        item = new Item(batchCat.cItem);
        holding = new Holding();
    }

    // ------------------------------------------------------------------------------- Main runloop --------

    public int Run (bool unbuffered)
    {
        Match m;
        try
        {
            // Read commands from standard input -- one command per command
            while (isRunning && (command = controlReader.ReadLine()) != null)
            {
                if (echo)
                    Console.Error.WriteLine(@"> {0}", command);

                // --- Ignore blank lines and comments
                if (command.Equals(@"") || (m = Regex.Match(command, @"^\s*(#.*)?$")).Success)
                { 
                    Console.WriteLine(@"IGNORED"); 
                }
                else if (command.Equals(@"blank"))
                {
                    Console.WriteLine(@"");
                }

                // --- Flow control
                else if ((m = Regex.Match(command, @"^begin (.+)$")).Success)
                {
                    HandleBegin(m.Groups[1].Value);
                }
                else if (command.Equals(@"end"))
                {
                    HandleEnd();
                }
                else if (command.Equals(@"quit"))
                {
                    HandleQuit();
                }

                // --- Debugging, tracing, etc.
                else if ((m = Regex.Match(command, @"^note (.+)$")).Success)
                {
                    HandleNote(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^echo (on|off)$")).Success)
                {
                    HandleEcho(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^strict (on|off)$")).Success)
                {
                    HandleStrict(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^debug (on|off)$")).Success)
                {
                    HandleDebug(m.Groups[1].Value);
                }
                else if (command.Equals(@"ping"))
                {
                    Console.WriteLine(@"ACK");
                }
                else if (inError > 0)
                {
                    SKIP(command);
                    Console.Out.Flush();
                    continue;
                }

                /* ---------------------------------------------------------------------------------- *\
                |                                                                                      |
                |  All commands from this point forward are skipped if we're recovering from an error  |
                |                                                                                      |
                \* ---------------------------------------------------------------------------------- */

                // --- Basic parameters used in connecting, etc.
                else if ((m = Regex.Match(command, @"^user (.+)$")).Success)
                {
                    HandleUser(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^password (.+)$")).Success)
                {
                    HandlePassword(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^inidir (.+)$")).Success)
                {
                    HandleIniDir(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^catloc (\d+)$")).Success)
                {
                    HandleCatalogingLocation(m.Groups[1].Value);
                }
                else if (command.Equals(@"connect"))
                {
                    if (userName != null && password != null && voyagerINIDir != null)
                        HandleConnect();
                    else
                        ERR(@"Can't connect: need username, password, and ini file");
                }

                // --- Operations on bib records
                else if ((m = Regex.Match(command, @"^bib id (\d+)$")).Success)
                {
                    HandleBibID(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^bib ownlib (\d+)$")).Success)
                {
                    HandleBibOwningLibrary(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^bib data :base64\((\d+)\) {$")).Success)
                {
                    HandleBibDataBase64(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^bib suppress (yes|no)$")).Success)
                {
                    HandleBibSuppress(m.Groups[1].Value);
                }
                else if (command.Equals(@"add bib"))
                {
                    HandleAddBib();
                }
                else if (command.Equals(@"replace bib"))
                {
                    HandleReplaceBib();
                }
                else if (command.Equals(@"delete bib"))
                {
                    HandleDeleteBib();
                }
                else if (command.Equals(@"print bib id"))
                {
                    HandlePrintBibID();
                }
                else if (command.Equals(@"print bib data"))
                {
                    HandlePrintBibData();
                }

                // --- Operations on holding records
                else if ((m = Regex.Match(command, @"^holding id (\d+)$")).Success)
                {
                    HandleHoldingID(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^holding data :base64\((\d+)\) {$")).Success)
                {
                    HandleHoldingDataBase64(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^holding suppress (yes|no)$")).Success)
                {
                    HandleHoldingSuppress(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^holding location (\d+)$")).Success)
                {
                    HandleHoldingLocation(m.Groups[1].Value);
                }
                else if (command.Equals(@"add holding"))
                {
                    HandleAddHolding();
                }
                else if (command.Equals(@"replace holding"))
                {
                    HandleReplaceHolding();
                }
                else if (command.Equals(@"delete holding"))
                {
                    HandleDeleteHolding();
                }

                // --- Operations on item records
                else if (command.Equals(@"clear item"))
                {
                    HandleClearItem();
                }
                else if ((m = Regex.Match(command, @"^item id (\d+)$")).Success)
                {
                    HandleItemID(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item type (\d+)$")).Success)
                {
                    HandleItemType(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item location (\d+)$")).Success)
                {
                    HandleItemLocation(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item caption (.+)$")).Success)
                {
                    HandleItemCaption(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item chron (.+)$")).Success)
                {
                    HandleItemChronology(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item copy (\d+)$")).Success)
                {
                    HandleItemCopyNumber(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item enum (.+)$")).Success)
                {
                    HandleItemEnumeration(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item free text (.+)$")).Success)
                {
                    HandleItemFreeText(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item media type (\d+)$")).Success)
                {
                    HandleItemMediaType(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item pieces (\d+)$")).Success)
                {
                    HandleItemPieces(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item price (.+)$")).Success)
                {
                    HandleItemPrice(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item spine label (.+)$")).Success)
                {
                    HandleItemSpineLabel(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item temp location (\d+)$")).Success)
                {
                    HandleItemTempLocation(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item temp type (\d+)$")).Success)
                {
                    HandleItemTempType(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^item year (.+)$")).Success)
                {
                    HandleItemYear(m.Groups[1].Value);
                }
                else if ((m = Regex.Match(command, @"^barcode (.+)$")).Success)
                {
                    HandleBarcode(m.Groups[1].Value);
                }
                else if (command.Equals(@"add item"))
                {
                    HandleAddItem();
                }
                else if (command.Equals(@"add barcode"))
                {
                    HandleAddBarcode();
                }
                else if (command.Equals(@"replace item"))
                {
                    HandleReplaceItem();
                }
                else if (command.Equals(@"delete item"))
                {
                    HandleDeleteItem();
                }

                // --- Operations on input files
                else if ((m = Regex.Match(command, @"^open input (.+)$")).Success)
                {
                    HandleOpenInput(m.Groups[1].Value);
                }
                else if (command.Equals(@"close input"))
                {
                    HandleCloseInput();
                }
                else if ((m = Regex.Match(command, @"^seek (\d+)$")).Success)
                {
                    HandleSeek(m.Groups[1].Value);
                }
                else if (command.Equals(@"read"))
                {
                    HandleRead();
                }

                // --- Oops!  No such command
                else
                {
                    ERR(String.Format(@"Unrecognized command: {0}", command));
                }
                Console.Out.Flush();
            }
        }
        catch(Exception e)
        {
            if (!e.Message.Equals(@""))
            {
                FATAL(String.Format(@"Unhandled exception: {0}", e.Message));
                // throw e;
            }
            return 1;
        }
        // End of runloop
        return 0;
    }
    
    // ------------------------------------------------------------------------------- Command handlers --------

    // Connecting...

    private void HandleUser(string argName)
    {
        userName = argName;
        OK();
    }
    private void HandlePassword (string argPassword)
    {
        password = argPassword;
        OK();
    }
    private void HandleIniDir (string argDir)
    {
        voyagerINIDir = argDir;
        OK();
    }
    private void HandleCatalogingLocation (string argLoc)
    {
        catalogingLocation = Convert.ToInt32(argLoc);
        OK();
    }
    private void HandleConnect ()
    {
        ConnectReturnCodes result = batchCat.Connect(ref voyagerINIDir, ref userName, ref password);
        if (result == ConnectReturnCodes.crSuccess)
        {
            isConnected = true;
            OK();
        }
        else
        {
            ERR(String.Format(@"Can't connect: {0}", result));
        }
    }

    // Operations on bib records

    private void HandleBibID (string argID)
    {
        bib.Id(argID);
        OK();
    }
    private void HandleBibOwningLibrary (string argOwnLib)
    {
        bib.owningLibrary = Convert.ToInt32(argOwnLib);
        OK();
    }
    private void HandleBibDataBase64 (string argDecodedLength)
    {
        int decodedLength = Convert.ToInt32(argDecodedLength);
        string str = ReadBase64Data(decodedLength);
        if (str == null)
        {
            ERR(@"Invalid Base64 bib data");
        }
        else
        {
            bib.Data(str);
            OK();
        }
    }
    private void HandleBibSuppress (string argToggle)
    {
        bib.Suppress(argToggle);
        OK();
    }
    private void HandleAddBib ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        AddBibReturnCode result;
        result = batchCat.AddBibRecord(ref bib.data, ref bib.owningLibrary, ref catalogingLocation, ref bib.suppress);
        if (result == AddBibReturnCode.abSuccess) {
            bib.Id(batchCat.RecordIDAdded);
            OK(Convert.ToString(batchCat.RecordIDAdded));
        }
        else
        {
            ERR(String.Format(@"Can't add bib: {0}", result));
        }
    }
    private void HandleReplaceBib ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        DateTime now = DateTime.Now;
        UpdateBibReturnCode result;
        result = batchCat.UpdateBibRecord(ref bib.id, ref bib.data, ref now, ref bib.owningLibrary, ref catalogingLocation, ref bib.suppress);
        if (result == UpdateBibReturnCode.ubSuccess)
        {
            OK();
        }
        else
        {
            ERR(String.Format(@"Can't replace bib ID {0}: {1}", bib.id, result));
        }
    }
    private void HandleDeleteBib ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        DeleteBibReturnCode result;
        result = batchCat.DeleteBibRecord(ref bib.id);
        if (result == DeleteBibReturnCode.dbSuccess)
            OK();
        else
            ERR(String.Format(@"Can't delete bib id {0}: {1}", bib.id, result));
    }

    // Operations on holding records

    private void HandleHoldingID (string argID)
    {
        holding.Id(argID);
        item.HoldingID(argID);
        OK();
    }
    private void HandleHoldingDataBase64 (string argDecodedLength)
    {
        int decodedLength = Convert.ToInt32(argDecodedLength);
        string str = ReadBase64Data(decodedLength);
        if (str == null)
        {
            ERR(@"Invalid Base64 holding data");
        }
        else
        {
            holding.Data(str);
            OK();
        }
    }
    private void HandleHoldingSuppress (string argToggle)
    {
        holding.Suppress(argToggle);
        OK();
    }
    private void HandleHoldingLocation (string argLocation)
    {
        holding.Location(argLocation);
        OK();
    }
    private void HandleAddHolding ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        AddHoldingReturnCode result;
        result = batchCat.AddHoldingRecord(ref holding.data, ref bib.id, ref catalogingLocation, ref holding.suppress, ref holding.location);
        if (result == AddHoldingReturnCode.ahSuccess) {
            holding.Id(batchCat.RecordIDAdded);
            OK(Convert.ToString(batchCat.RecordIDAdded));
        }
        else
        {
            ERR(String.Format(@"Can't add holding: {0}", result));
        }
    }
    private void HandleReplaceHolding ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        DateTime now = DateTime.Now;
        UpdateHoldingReturnCode result;
        result = batchCat.UpdateHoldingRecord(ref holding.id, ref holding.data, ref now, ref catalogingLocation, ref bib.id, ref holding.location, ref holding.suppress);
        if (result == UpdateHoldingReturnCode.uhSuccess)
            OK();
        else
            ERR(String.Format(@"Can't replace holding ID {0}: {1}", holding.id, result));
    }
    private void HandleDeleteHolding ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        DeleteHoldingReturnCode result;
        result = batchCat.DeleteHoldingRecord(ref holding.id);
        if (result == DeleteHoldingReturnCode.dhSuccess)
            OK();
        else
            ERR(String.Format(@"Can't delete holding id {0}: {1}", holding.id, result));
    }

    // Operations on items

    private void HandleClearItem ()
    {
        item.Clear();
        OK();
    }
    private void HandleItemID (string argID)
    {
        item.Id(argID);
        OK();
    }
    private void HandleItemType (string argID)
    {
        item.Type(argID);
        OK();
    }
    private void HandleItemLocation (string argLocation)
    {
        item.PermLocation(argLocation);
        OK();
    }
    private void HandleItemCaption(string argCaption)
    {
        item.Caption(argCaption);
        OK();
    }
    private void HandleItemChronology(string argChronology)
    {
        item.Chronology(argChronology);
        OK();
    }
    private void HandleItemCopyNumber(string argCopyNumber)
    {
        item.CopyNumber(argCopyNumber);
        OK();
    }
    private void HandleItemEnumeration(string argEnumeration)
    {
        item.Enumeration(argEnumeration);
        OK();
    }
    private void HandleItemFreeText(string argFreeText)
    {
        item.FreeText(argFreeText);
        OK();
    }
    private void HandleItemMediaType(string argMediaType)
    {
        item.MediaType(argMediaType);
        OK();
    }
    private void HandleItemPieces(string argPieces)
    {
        item.Pieces(argPieces);
        OK();
    }
    private void HandleItemPrice(string argPrice)
    {
        item.Price(argPrice);
        OK();
    }
    private void HandleItemSpineLabel(string argSpineLabel)
    {
        item.SpineLabel(argSpineLabel);
        OK();
    }
    private void HandleItemTempLocation(string argLocation)
    {
        item.TempLocation(argLocation);
        OK();
    }
    private void HandleItemTempType(string argID)
    {
        item.TempType(argID);
        OK();
    }
    private void HandleItemYear(string argYear)
    {
        item.Year(argYear);
        OK();
    }

    private void HandleBarcode (string argBarcode)
    {
        item.Barcode(argBarcode);
        OK();
    }
    private void HandleAddItem ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        AddItemReturnCode result;
        item.HoldingId(holding.id);
        result = batchCat.AddItemData(ref catalogingLocation);
        if (result == AddItemReturnCode.aiSuccess)
        {
            item.Id(batchCat.RecordIDAdded);
            OK(Convert.ToString(batchCat.RecordIDAdded));
        }
        else
        {
            ERR(String.Format(@"Can't add item: {0}", result));
        }
    }
    private void HandleAddBarcode ()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        AddItemBarCodeReturnCode result;
        result = batchCat.AddItemBarCode(ref item.id, ref item.barcode);
        if (result == AddItemBarCodeReturnCode.aibSuccess)
            OK();
        else
            ERR(String.Format(@"Can't add barcode to item {0}: {1}", item.id, result));
    }
    private void HandleReplaceItem()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        short result;
        result = batchCat.UpdateItemData(ref catalogingLocation);
        if (result == 0)
            OK();
        else
            ERR(String.Format(@"Can't replace item ID {0}: {1}", item.id, result));
    }
    private void HandleDeleteItem()
    {
        if (!isConnected)
            FATAL(@"Not connected");
        DeleteItemReturnCode result;
        result = batchCat.DeleteItemRecord(ref item.id);
        if (result == DeleteItemReturnCode.diSuccess)
            OK();
        else
            ERR(String.Format(@"Can't delete item id {0}: {1}", item.id, result));
    }

    // Reading input...

    private void HandleOpenInput (string argFile)
    {
        inputFile = argFile;
        inputReader = OpenInputFile(inputFile);
        if (inputReader != null)
            OK();
        else
            ERR(@"Can't open input file");
    }
    private void HandleCloseInput ()
    {
        inputReader.Close();
        inputReader = null;
        OK();
    }
    private void HandleSeek (string argOffset)
    {
        if (inputReader != null)
        {
            long offset = Convert.ToInt32(argOffset);
            inputReader.BaseStream.Seek(offset, SeekOrigin.Begin);
            OK();
        }
        else
        {
            ERR(@"No open file");
        }
    }
    private void HandleRead ()
    {
        byte[]  lengthAsBytes;
        int     length;
        byte[]  recordAsBytes;
        lengthAsBytes = new byte[5];
        inputReader.Read(lengthAsBytes, 0, 5);
        length = Convert.ToInt32(utf8.GetString(lengthAsBytes));
        recordAsBytes = new byte[length];
        for (int i = 0; i < 5; i++)
            recordAsBytes[i] = lengthAsBytes[i];
        inputReader.Read(recordAsBytes, 5, length - 5);
        bib.Data(BytesToRecord(recordAsBytes));
        OK();
    }

    // Other...
    private void HandlePrintBibID ()
    {
        VALUE(Convert.ToString(bib.id));
    }
    private void HandlePrintBibData()
    {
        VALUE(Convert.ToString(bib.data));
    }
    private void HandleBegin(string argNote)
    {
        Console.WriteLine(@"BEGIN {0}", argNote);
        blockLevel++;
    }
    private void HandleEnd()
    {
        if (inError > 0)
        {
            inError--;
            blockLevel--;
            if (inError == 0)
                Console.WriteLine(@"END recovery: done, returning to normal state");
            else
                Console.WriteLine(@"END recovery: popping to error level {0}", blockLevel);
        }
        else if (blockLevel > 0)
        {
            Console.WriteLine(@"END OK");
            blockLevel--;
        }
        else
        {
            ERR(@"end without begin");
        }
    }
    private void HandleNote(string argNote)
    {
        Console.WriteLine(@"NOTE {0}", argNote);
    }
    private void HandleEcho(string argToggle)
    {
        echo = (argToggle.Equals(@"on"));
        OK();
    }
    private void HandleStrict(string argToggle)
    {
        strict = (argToggle.Equals(@"on"));
        OK();
    }
    private void HandleDebug(string argToggle)
    {
        debug = (argToggle.Equals(@"on"));
        OK();
    }
    private void HandleQuit()
    {
        isRunning = false;
        OK();
    }

    // ------------------------------------------------------------------------------- Utility functions --------

    private BinaryReader OpenInputFile (string filename)
    {
        BinaryReader reader;
        if (!File.Exists(filename))
            return null;
        reader = new BinaryReader(File.Open(filename, FileMode.Open));
        return reader;
    }

    private string ReadBase64Data (int decodedLength)
    {
        byte[] recordAsBytes = new byte[decodedLength];
        int lengthSoFar = 0;
        string line;
        try
        {
            while ((line = controlReader.ReadLine()) != null)
            {
                Console.WriteLine(@"DATA");
                if (line.Equals(@"}"))
                    break;
                byte[] lineDecoded = Convert.FromBase64String(line);
                lineDecoded.CopyTo(recordAsBytes, lengthSoFar);
                lengthSoFar += lineDecoded.Length;
                if (lengthSoFar > decodedLength)
                    return null;
            }
            if (lengthSoFar < decodedLength)
                return null;
        }
        catch
        {
            return null;
        }
        return BytesToRecord(recordAsBytes);
    }

    private string BytesToRecord(byte[] bytes)
    {
        // return utf8.GetString(bytes);
        return EncodeForBatchCat(utf8.GetString(bytes));
    }

    /// <summary>
    /// Apparently, BatchCat runs each character through the Windows codepage;
    /// this function is a workaround that converts a MARC record to the form that
    /// BatchCat requires.  Adapted from code by Reuben Pasquini.
    /// </summary>
    /// <param name="s">what the world thinks is a string</param>
    /// <returns>what BatchCat thinks is a string</returns>
    private string EncodeForBatchCat(string s)
    {
        // --- This is what I would have expected to work:
        // Encoding encoding = utf8;
        // return new String(utf8.GetBytes(s))

        // --- This is what Reuben Pasquini seems to be suggesting:
        Encoding encoding = Encoding.Default;
        return new String(encoding.GetChars(utf8.GetBytes(s)));
    }

    // ------------------------------------------------------------------------------- Results functions --------

    private void OK()
    {
        Console.WriteLine(@"OK");
    }
    private void OK(string result)
    {
        Console.WriteLine(@"OK {0}", result);
    }
    private void ERR (string msg)
    {
        if (strict)
            FATAL(msg);
        else if (blockLevel > 0)
            inError = blockLevel;
        else
            FATAL(String.Format(@"Error outside of protected block: {0}", msg));
        Console.WriteLine(@"ERR {0}", msg);
    }
    private void FATAL(string msg)
    {
        Console.WriteLine(@"FATAL {0}", msg);
        throw new Exception(@"");
    }
    private void SKIP(string cmd)
    {
        if (echo)
            Console.WriteLine(@"SKIP | {0}", command);
        else
            Console.WriteLine(@"SKIP");
    }
    private void VALUE (string val)
    {
        Console.WriteLine(@"VALUE {0}", val);
    }
    private void DEBUG (string msg)
    {
        if (debug)
            Console.Error.WriteLine(@"# {0}", msg);
    }

/*
    private void TODO (string message)
    {
        Console.WriteLine("< TODO {0}", message);
    }
    public static String ReadMARCRecord (Stream stream, ref long bibid)
    {
        DEBUG(">> ReadMARCRecord");
        // Read the record length (first 5 bytes of the record)
        byte[] reclenstr = new byte[5];
        int reclen;
        try
        {
            reclen = ReadInt(stream, reclenstr);
        }
        catch (EndOfStreamException exception)
        {
            return null;
        }
        // Allocate memory for the entire record
        DEBUG("reclen " + reclen);
        byte[] record = new byte[reclen];
        // Copy the first five bytes to the allocated record
        reclenstr.CopyTo(record, 0);
        // Find the 001 field contents
        int baseaddr   = GetInt(record, 12, 5);
        DEBUG("baseaddr " + baseaddr);
        int f001offset = GetInt(record, baseaddr + 7, 5);
        int f001len    = GetInt(record, baseaddr + 3, 4);
        byte[] f001    = GetBytes(record, f001offset, f001len-1);
        bibid = IntFromBytes(f001);
        DEBUG("<< ReadMARCRecord");
        return Encoding.UTF8.GetString(record);
    }

    public static byte[] GetBytes (byte[] record, int offset, int len)
    {
        byte[] bytes = new byte[len];
        for (int i = 0; i < len; i++) {
            record[i] = record[offset + i];
        }
        return bytes;
    }

    public static int GetInt (byte[] record, int offset, int length)
    {
        // DEBUG(String.Format(">> GetInt (offset: {0}, length: {1})", offset, length));
        byte[] buffer = new byte[length];
        for (int i = 0; i < length; i++) {
            // Console.WriteLine("{0} ", byteToHexString(buffer[i]));
            buffer[i] = record[offset+i];
        }
        // Console.Write("\n");
        // Console.WriteLine("<< GetInt");
        return IntFromBytes(buffer);
    }

    public static int ReadInt (Stream stream, byte[] bytes)
    {
        FillBytes(stream, bytes);
        return IntFromBytes(bytes);
    }

    public static void FillBytes (Stream stream, byte[] data)
    {
        int offset = 0;
        int remaining = data.Length;
        while (remaining > 0)
            {
                int read = stream.Read(data, offset, remaining);
                if (read <= 0)
                    throw new EndOfStreamException(String.Format("End of stream reached with {0} bytes left to read", remaining));
                remaining -= read;
                offset += read;
            }
    }

    public static int IntFromBytes (byte[] bytes)
    {
        // Console.WriteLine(">> IntFromBytes");
        foreach (byte b in bytes)
        {
            Console.Write("{0} ", byteToHexString(b));
        }
        Console.Write("\n");
        int value = 0;
        foreach (byte digit in bytes)
        {
            if (digit < 0x30 || digit > 0x39)
                throw new Exception();
            value *= 10;
            value += digit - 0x30;
        }
        // Console.WriteLine("<< IntFromBytes");
        return value;
    }

    private static string byteToHexString (byte b)
    {
        return padString(String.Format("{0}", b.ToString("X")));
    }

    private static string padString (string s)
    {
        while (s.Length < 2) s = "0" + s;
        return s;
    }
*/

}


